'''

transfer methods

'''
import os
import traceback
import importlib.machinery
from komlogd.base import config, logging
from komlogd.base.settings import options

def load_transfer_methods_files():
    entries = config.config.get_entries(entryname=options.ENTRY_TRANSFERS)
    files=[]
    for entry in entries:
        try:
            enabled=entry[options.TRANSFERS_ENABLED]
            filename=entry[options.TRANSFERS_FILE]
            if enabled is True:
                logging.logger.debug('adding transfer methods file: '+str(filename))
                files.append(filename)
        except Exception:
            logging.logger.error('Error loading transfer methods configuration.')
            ex_info=traceback.format_exc().splitlines()
            for line in ex_info:
                logging.logger.error(line)
    if len(files) == 0:
        logging.logger.info('No transfer methods loaded.')
    for i,item in enumerate(files):
        if not os.path.isabs(item):
            path=os.path.join(config.config.root_dir,item)
        else:
            path=item
        try:
            importlib.machinery.SourceFileLoader(fullname='dynimp'+str(i),path=path).load_module()
        except Exception:
            logging.logger.error('Error loading transfer methods file '+str(item))
            ex_info=traceback.format_exc().splitlines()
            for line in ex_info:
                logging.logger.error(line)




'''
Default values for komlogd configuration parameters

'''

APP_PATH = '.komlogd'
CONFIG_FILE = 'komlogd.yaml'
LOG_FILE='komlogd.log'
LOG_DIR='log/'
LOG_LEVEL='INFO'
LOG_ROTATION='yes'
LOG_MAX_BYTES=10000000
LOG_BACKUP_COUNT=3
LOG_FORMAT='%(asctime)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s'
RSA_PRIV_KEY = 'key.priv'
RSA_PUB_KEY = 'key.pub'




import asyncio
import aiohttp
import json
import traceback
import time
import pandas as pd
from komlogd.api import logging, exceptions, crypto
from komlogd.api.protocol.model import messages, validation
from komlogd.api.protocol.processing import message as prmsg
from komlogd.api.protocol.processing import procedure as prproc
from komlogd.api.model import store, transfer_methods, queues

KOMLOG_LOGIN_URL = 'https://www.komlog.io/login'
KOMLOG_WS_URL = 'https://agents.komlog.io/'


class KomlogSession:
    def __init__(self, username, privkey, loop=None, load_tm=True):
        loop = loop or asyncio.get_event_loop()
        self._loop = loop
        self.username = username
        self.privkey = privkey
        self._load_tm = load_tm
        self._metrics_store = store.MetricsStore(owner=self.username)
        self._transfer_methods = transfer_methods.TransferMethodsIndex(owner=self.username)
        self._session = None
        self._ws = None
        self._session_future = None
        self._loop_future = None
        self._deferred = []
        self._waiting_response = {}
        self._q_msg_workers = queues.AsyncQueue(num_workers=5, on_msg=self._process_received_message, name='Message Workers', loop=self._loop)

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        try:
            validation.validate_username(value)
            getattr(self, '_username')
            raise exceptions.BadParametersException('username modification not allowed')
        except AttributeError:
            self._username = value
        except TypeError:
            raise exceptions.BadParametersException('Invalid username {}'.format(str(value)))

    @property
    def privkey(self):
        return self._privkey

    @privkey.setter
    def privkey(self, value):
        try:
            validation.validate_privkey(value)
            getattr(self, '_privkey')
            raise exceptions.BadParametersException('private key modification not allowed')
        except AttributeError:
            self._privkey = value
            self._pubkey = value.public_key()
            self._serialized_pubkey = crypto.serialize_public_key(self._pubkey)
            self._printable_pubkey = crypto.get_printable_pubkey(self._pubkey)
        except TypeError:
            raise exceptions.BadParametersException('Invalid private key')

    async def close(self):
        logging.logger.info('closing Komlog connection')
        self._stop_f = True
        await self._q_msg_workers.join()
        if self._ws and self._ws.closed is False:
            await self._ws.close()
        if self._session:
            await self._session.close()
        if self._loop_future:
            await self._loop_future
            self._session_future.set_result(True)

    async def login(self):
        if self._session_future is None:
            logging.logger.info('Authenticating agent')
            await self._auth()
            logging.logger.info('Initializing websocket connection')
            await self._ws_connect()
            self._q_msg_workers.start()
            if self._load_tm:
                self._load_anonymous_transfer_methods()
            logging.logger.info('Entering loop')
            self._loop_future = asyncio.ensure_future(self._session_loop(), loop=self._loop)
            self._session_future = asyncio.futures.Future(loop=self._loop)

    async def join(self):
        if self._session_future is not None:
            while not getattr(self, '_stop_f',False):
                await self._session_future

    async def _auth(self):
        self._session = aiohttp.ClientSession()
        data = {
            'u':self._username,
            'k':self._serialized_pubkey,
            'pv':messages.KomlogMessage._version_
        }
        try:
            async with self._session.post(KOMLOG_LOGIN_URL,data=data) as resp:
                resp_content = await resp.json()
                if resp.status == 403:
                    logging.logger.error('Access Denied')
                    logging.logger.error('is username correct? '+self._username)
                    logging.logger.error('is agent public key added on web and in active state?')
                    logging.logger.error('public key content is:\n'+self._printable_pubkey)
                    raise exceptions.LoginException('Access denied')
                elif not (resp.status == 200 and 'challenge' in resp_content):
                    logging.logger.error('Unexpected server response: '+str(resp))
                    raise exceptions.LoginException('Unexpected error')
            c = crypto.process_challenge(self._privkey, resp_content['challenge'])
            s = crypto.sign_message(self._privkey, c)
            data['c']=c
            data['s']=s
            async with self._session.post(KOMLOG_LOGIN_URL,data=data) as resp:
                resp_content = await resp.json()
                if resp.status == 403:
                    logging.logger.error('Access Denied. is agent active?')
                    raise exceptions.LoginException('Authentication process failed')
        except:
            if self._session:
                await self._session.close()
            raise

    async def _ws_connect(self):
        try:
            self._ws = await self._session.ws_connect(KOMLOG_WS_URL)
        except:
            if self._ws:
                await self._ws.close()
            raise

    async def _session_loop(self):
        while not getattr(self, '_stop_f',False):
            try:
                if not self._session:
                    logging.logger.debug('Restarting Komlog session')
                    await self._auth()
                    await self._ws_connect()
                elif not self._ws:
                    logging.logger.debug('Restarting websocket connection')
                    await self._ws_connect()
                self._ws_reconnected()
                async for msg in self._ws:
                    logging.logger.debug('Message received from server: '+str(msg))
                    if msg.tp == aiohttp.WSMsgType.CLOSED:
                        break
                    elif msg.tp == aiohttp.WSMsgType.ERROR:
                        break
                    else:
                        await self._q_msg_workers.push(msg)
                        self._metrics_store.run_maintenance()
            except Exception:
                ex_info=traceback.format_exc().splitlines()
                for line in ex_info:
                    logging.logger.error(line)
            finally:
                logging.logger.debug('Unexpected session close')
                if self._ws and self._ws.closed:
                    if self._ws.close_code == 4403:
                        logging.logger.debug('Server denied access. Retrying connection.')
                        self._session = None
                    self._ws = None
                if not getattr(self, '_stop_f',False):
                    self._ws_disconnected()
                    await asyncio.sleep(15)

    def _load_anonymous_transfer_methods(self):
        logging.logger.debug('Loading transfer methods')
        for item in transfer_methods.anon_transfer_methods.get_transfer_methods(enabled=False):
            if not self._transfer_methods.add_transfer_method(item, enabled=False):
                logging.logger.error('Error loading transfer method '+item.f.__name__)

    def _ws_disconnected(self):
        self._transfer_methods.disable_all()

    def _ws_reconnected(self):
        for msg in self._deferred[:]:
            logging.logger.debug('sending deferred message')
            self._deferred.remove(msg)
            self._send_message(msg)
        asyncio.ensure_future(prproc.initialize_transfer_methods(self))

    async def _periodic_transfer_method_call(self, mid):
        t = time.time()
        await asyncio.sleep(60-t%60)
        localtime = time.localtime(t+(60-t%60))
        tm_info = self._transfer_methods.get_transfer_method_info(mid)
        if (tm_info and
            tm_info['enabled']):
            if tm_info['tm'].schedule.meets(t=localtime):
                ts=pd.Timestamp(ts_input=t+(60-t%60), unit='s', tz='utc')
                await prproc.exec_transfer_method(session=self, mid=tm_info['tm'].mid, ts=ts, metrics=[])
            asyncio.ensure_future(self._periodic_transfer_method_call(mid))

    async def _process_received_message(self, msg):
        try:
            data=json.loads(msg.data)
            if 'action' in data:
                message=messages.KomlogMessage.load_from_dict(data)
                if message.irt and message.irt in self._waiting_response:
                    if self._waiting_response[message.irt].done():
                        # some messages generate multiple responses. Protocol procedures are responsible for
                        # marking msg done or undone, adding more pending futures in the last case.
                        logging.logger.debug('Enqueueing message again')
                        await self._q_msg_workers.push(msg)
                    else:
                        logging.logger.debug('processing message response procedure')
                        self._waiting_response[message.irt].set_result(message)
                else:
                    logging.logger.debug('processing non requested message')
                    prmsg.processing_map[message.action](msg=message, session=self)
        except Exception:
            ex_info=traceback.format_exc().splitlines()
            for line in ex_info:
                logging.logger.error(line)

    def _send_message(self, message):
        if not isinstance(message, messages.KomlogMessage):
            raise exceptions.InvalidMessageException()
        try:
            logging.logger.debug('sending message '+str(message.to_dict()))
            self._ws.send_str(json.dumps(message.to_dict()))
        except Exception:
            ex_info=traceback.format_exc().splitlines()
            for line in ex_info:
                logging.logger.error(line)
            self._deferred.append(message)

    async def _await_response(self, message):
        self._send_message(message)
        future = self._mark_message_undone(message.seq)
        return await future

    def _mark_message_done(self, seq):
        self._waiting_response.pop(seq,None)

    def _mark_message_undone(self, seq):
        future = asyncio.futures.Future(loop=self._loop)
        self._waiting_response[seq]=future
        return future

    async def send_samples(self, samples):
        logging.logger.debug('sending samples')
        result = await prproc.send_samples(self, samples)
        logging.logger.debug('result: {}'.format(str(result)))
        return result




import pandas as pd

class DataRequirements:
    def __init__(self, past_delta=None, past_count=None):
        self.past_delta = past_delta
        self.past_count = past_count

    @property
    def past_delta(self):
        return self._past_delta

    @past_delta.setter
    def past_delta(self, value):
        if isinstance(value, pd.Timedelta) or value is None:
            self._past_delta = value
        else:
            raise TypeError('Invalid past_delta parameter')

    @property
    def past_count(self):
        return self._past_count

    @past_count.setter
    def past_count(self, value):
        if (isinstance(value, int) and value >= 0) or value is None:
            self._past_count = value
        else:
            raise TypeError('Invalid past_count parameter')





'''

Datastores

'''

import time
import decimal
import pandas as pd
from komlogd.api import logging, uri
from komlogd.api.protocol.model import validation
from komlogd.api.protocol.model.transfer_methods import DataRequirements
from komlogd.api.protocol.model.types import Metric, Datasource, Datapoint


class MetricsStore:
    def __init__(self, owner, maintenance_exec_delta=None, data_delta=None):
        self.owner = owner
        self.maintenance_exec_delta=maintenance_exec_delta
        self.default_reqs=data_delta
        self._last_maintenance_exec=pd.Timestamp('now',tz='utc')
        self._metric_reqs={}
        self._series={}

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        validation.validate_username(value)
        self._owner = value.lower()

    @property
    def maintenance_exec_delta(self):
        return self._maintenance_exec_delta

    @maintenance_exec_delta.setter
    def maintenance_exec_delta(self, value):
        if value is None:
            self._maintenance_exec_delta=pd.Timedelta('5 m')
        elif isinstance(value, pd.Timedelta):
            self._maintenance_exec_delta=value
        else:
            raise TypeError('Invalid maintenance_exec_delta parameter')

    @property
    def default_reqs(self):
        return self._default_reqs

    @default_reqs.setter
    def default_reqs(self, value):
        if value is None:
            self._default_reqs=DataRequirements(past_delta=pd.Timedelta('10min'),past_count=2)
        elif isinstance(value, DataRequirements):
            self._default_reqs=value
        else:
            raise TypeError('Invalid data_delta parameter')

    def store(self, metric, ts, content):
        if isinstance(metric, Datasource):
            validation.validate_ds_content(content)
        elif isinstance(metric, Datapoint):
            validation.validate_dp_content(content)
        else:
            return
        validation.validate_ts(ts)
        ts=pd.Timestamp(ts.astimezone('utc'))
        guri = uri.get_global_uri(metric, owner=self.owner)
        try:
            if isinstance(content, decimal.Decimal):
                value = int(content) if content%1 == 0 else float(content)
            else:
                value = content
            self._series[guri][ts]=value
            if self._series[guri].index[-1]<self._series[guri].index[-2]:
                self._series[guri]=self._series[guri].sort_index()
        except KeyError:
            self._series[guri]=pd.Series(data=[value], index=[ts])
        except IndexError:
            pass

    def isin(self, metric, ts, content):
        guri = uri.get_global_uri(metric, owner=self.owner)
        if not guri in self._series:
            return False
        elif not ts in self._series[guri]:
            return False
        elif not self._series[guri][ts] == content:
            return False
        return True

    def run_maintenance(self):
        now = pd.Timestamp('now',tz='utc')
        if now - self.maintenance_exec_delta > self._last_maintenance_exec:
            self.purge()
            self._last_maintenance_exec=pd.Timestamp('now',tz='utc')

    def purge(self):
        now = pd.Timestamp('now',tz='utc')
        for guri in self._series.keys():
            reqs = self._metric_reqs[guri] if guri in self._metric_reqs else self._default_reqs
            past_delta = reqs.past_delta if reqs.past_delta else self._default_reqs.past_delta
            past_count = reqs.past_count if reqs.past_count else self._default_reqs.past_count
            t1_delta = now-past_delta if past_delta else None
            t1_count = self._series[guri][:now][-past_count:].index[0] if past_count else None
            if t1_delta or t1_count:
                count=self._series[guri].count()
                if t1_delta and t1_count:
                    t1 = t1_delta if t1_delta < t1_count else t1_count
                else:
                    t1 = t1_delta if t1_delta else t1_count
                self._series[guri]=self._series[guri].ix[t1:now]
                new_count=self._series[guri].count()
                logging.logger.debug('Deleted '+str(count-new_count)+' rows from time series '+guri+' before: '+str(count)+' now: '+str(new_count))
        end = pd.Timestamp('now',tz='utc')
        elapsed = end-now
        logging.logger.debug('Purge procedure finished. Elapsed time (s): '+'.'.join((str(elapsed.seconds),str(elapsed.microseconds).zfill(6))))
        return True

    def add_metric_data_reqs(self, metric, reqs):
        if not isinstance(metric, Metric):
            raise TypeError('Invalid metric type')
        if not isinstance(reqs, DataRequirements):
            raise TypeError('Invalid requirements type')
        guri = uri.get_global_uri(metric, owner=self.owner)
        if not guri in self._metric_reqs:
            self._metric_reqs[guri]=reqs
        else:
            new_reqs = DataRequirements()
            if reqs.past_delta and self._metric_reqs[guri].past_delta:
                new_reqs.past_delta = reqs.past_delta if reqs.past_delta > self._metric_reqs[guri].past_delta else self._metric_reqs[guri].past_delta
            elif reqs.past_delta:
                new_reqs.past_delta = reqs.past_delta
            else:
                new_reqs.past_delta = self._metric_reqs[guri].past_delta
            if reqs.past_count != None and self._metric_reqs[guri].past_count != None:
                new_reqs.past_count = reqs.past_count if reqs.past_count > self._metric_reqs[guri].past_count else self._metric_reqs[guri].past_count
            elif reqs.past_count != None:
                new_reqs.past_count = reqs.past_count
            else:
                new_reqs.past_count = self._metric_reqs[guri].past_count
            self._metric_reqs[guri] = new_reqs
        return True

    def get_metric_data_reqs(self, metric):
        guri = uri.get_global_uri(metric, owner=self.owner)
        try:
            return self._metric_reqs[guri]
        except KeyError:
            return None

    def get_serie(self, metric, ets=None, its=None, count=None):
        guri = uri.get_global_uri(metric, owner=self.owner)
        try:
            serie = self._series[guri]
            if not ets:
                d = serie.copy(deep=True)
            elif its:
                d = serie[its:ets].copy(deep=True)
            elif count:
                d = serie[:ets].tail(count).copy(deep=True)
            else:
                d = serie[:ets].tail(1).copy(deep=True)
            d.name = metric
            return d
        except KeyError:
            serie = pd.Series()
            serie.name = metric
            return serie




import os
import traceback
from base64 import b64encode, b64decode
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from komlogd.api import exceptions, logging

def load_private_key(privkey_file):
    ''' 
    Loads the private key stored in the file indicated by the privkey_file
    parameter and returns it

    The `privkey_file` parameter indicates the absolute path to the file
    storing the private key.

    The key returned is a RSAPrivateKey instance.

    If there is a problem accessing the file, the exception will be raised
    without modification.

    If the problem occurs loading the key, a CryptoException will be raised.
    '''
    with open(privkey_file, "rb") as key_file:
        try:
            privkey = serialization.load_pem_private_key(
                key_file.read(),
                password=None,
                backend=default_backend()
            )
            if isinstance(privkey, rsa.RSAPrivateKey):
                return privkey
        except Exception as e:
            raise exceptions.CryptoException()

def load_public_key(pubkey_file):
    ''' 
    Loads the public key stored in the file indicated by the pubkey_file
    parameter and returns it

    The `pubkey_file` parameter indicates the absolute path to the file
    storing the public key.

    The key returned is a RSAPublicKey instance.

    If there is a problem accessing the file, the exception will be raised
    without modification.

    If the problem occurs loading the key, a CryptoException will be raised.
    '''
    with open(pubkey_file, "rb") as key_file:
        try:
            pubkey = serialization.load_pem_public_key(
                key_file.read(),
                backend=default_backend()
            )
            if isinstance(pubkey, rsa.RSAPublicKey):
                return pubkey
        except Exception as e:
            raise exceptions.CryptoException()

def serialize_public_key(key):
    '''
    Returns the public key serialization in base64
    '''
    try:
        pem = key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return b64encode(pem).decode()
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def serialize_private_key(key):
    '''
    Returns the private key serialization in base64
    '''
    try:
        pem = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
        return b64encode(pem).decode()
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def decrypt(privkey, ciphertext):
    try:
        plaintext = privkey.decrypt(
            ciphertext,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        return plaintext
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def encrypt(pubkey, plaintext):
    try:
        ciphertext= pubkey.encrypt(plaintext=plaintext,
            padding=padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None
            )
        )
        return ciphertext
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def sign_message(privkey, message):
    try:
        plaintext = b64decode(message.encode('utf-8'))
        signer = privkey.signer(
            padding.PSS(
                mgf=padding.MGF1(hashes.SHA256()),
                salt_length=padding.PSS.MAX_LENGTH
            ),
            hashes.SHA256()
        )
        signer.update(plaintext)
        return b64encode(signer.finalize()).decode('utf-8')
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def get_hash(message):
    try:
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        for i in range(0,30):
            digest.update(message)
        return digest.finalize()
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def process_challenge(privkey, challenge):
    try:
        ciphertext = b64decode(challenge.encode('utf-8'))
        plaintext = decrypt(privkey=privkey, ciphertext=ciphertext)
        hashed = get_hash(plaintext)
        return b64encode(hashed).decode('utf-8')
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        return None

def generate_rsa_key(key_size=4096):
    privkey = rsa.generate_private_key(
        public_exponent=65537,
        key_size=key_size,
        backend=default_backend()
    )
    return privkey

def store_keys(privkey, privkey_file, pubkey_file):
    if not isinstance(privkey,rsa.RSAPrivateKey):
        return False
    privkey_serial = privkey.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()
    )
    pubkey = privkey.public_key()
    pubkey_serial = pubkey.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    priv_stored=False
    try:
        with open(privkey_file,'wb') as privkey_out:
            privkey_out.write(privkey_serial)
            privkey_out.close()
            priv_stored=True
        with open(pubkey_file,'wb') as pubkey_out:
            pubkey_out.write(pubkey_serial)
            pubkey_out.close()
    except Exception:
        ex_info=traceback.format_exc().splitlines()
        for line in ex_info:
            logging.logger.error(line)
        if priv_stored:
            os.remove(privkey_file)
        return False
    return True

def get_printable_pubkey(pubkey):
    if not isinstance(pubkey, rsa.RSAPublicKey):
        return None
    pem = pubkey.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    return pem.decode()




import asyncio
import dateutil.tz
import sys
import pandas as pd
from komlogd.api import session
from komlogd.api.protocol.model.types import Datasource, Sample
from komlogd.base import crypto, config, logging, exceptions
from komlogd.base.settings import options

KomlogSession = None

def initialize_komlog_session(load_tm=True):
    global KomlogSession
    username = config.config.get_entries(entryname=options.KOMLOG_USERNAME)
    privkey = crypto.get_private_key()
    if len(username)==0:
        raise exceptions.BadParametersException('Set username in configuration file.')
    elif len(username)>1:
        raise exceptions.BadParametersException('More than one username found in configuration file. Keep only one.')
    KomlogSession = session.KomlogSession(username=username[0], privkey=privkey, load_tm=load_tm)

async def start_komlog_session():
    await KomlogSession.login()
    await KomlogSession.join()

async def start_komlogd_stdin_mode(uri):
    now = pd.Timestamp('now', tz=dateutil.tz.tzlocal())
    data = sys.stdin.read()
    sample = Sample(metric=Datasource(uri), ts=now, data=data)
    await KomlogSession.login()
    result = await KomlogSession.send_samples([sample])
    if not result['success']:
        sys.stderr.write('Error sending data to Komlog.\n')
        sys.stderr.write(result['error']+'\n')
    await KomlogSession.close()

async def stop_komlog_session():
    await KomlogSession.close()




from komlogd.api import exceptions
from komlogd.api.protocol.model.types import Metric

class Schedule:
    def __new__(cls, *args, **kwargs):
        if cls is Schedule:
            raise TypeError("Schedule base class may not be instantiated")
        return object.__new__(cls)

    def __init__(self, exec_on_load):
        self.exec_on_load = exec_on_load

    @property
    def exec_on_load(self):
        return self._exec_on_load

    @exec_on_load.setter
    def exec_on_load(self, value):
        self._exec_on_load = bool(value)

class DummySchedule(Schedule):
    def __init__(self, exec_on_load=False):
        super().__init__(exec_on_load=exec_on_load)

class OnUpdateSchedule(Schedule):
    def __init__(self, metrics=None, exec_on_load=False):
        self.metrics = metrics
        super().__init__(exec_on_load=exec_on_load)

    @property
    def metrics(self):
        return self._metrics

    @metrics.setter
    def metrics(self, value):
        if value is None:
            self._metrics = []
        elif isinstance(value,list):
            for m in value:
                if not isinstance(m, Metric):
                    raise exceptions.BadParametersException('Invalid metric: '+str(m))
            else:
                self._metrics = value
        else:
            raise exceptions.BadParametersException('"metrics" attribute must be a list')

class CronSchedule(Schedule):
    def __init__(self, minute='*', hour='*', month='*', dow='*', dom='*', exec_on_load=False):
        super().__init__(exec_on_load=exec_on_load)
        self._schedule={}
        self.minute = minute
        self.hour = hour
        self.month = month
        self.dow = dow
        self.dom = dom

    @property
    def minute(self):
        return self._minute

    @minute.setter
    def minute(self, value):
        sched_entry = self._process_var(value, 59, 0)
        if sched_entry:
            self._schedule['minute']=sched_entry
            self._minute = value
        else:
            raise exceptions.BadParametersException('Invalid minute value: '+str(value))

    @property
    def hour(self):
        return self._hour

    @hour.setter
    def hour(self, value):
        sched_entry = self._process_var(value, 23, 0)
        if sched_entry:
            self._schedule['hour']=sched_entry
            self._hour = value
        else:
            raise exceptions.BadParametersException('Invalid hour value: '+str(value))

    @property
    def month(self):
        return self._month

    @month.setter
    def month(self, value):
        sched_entry = self._process_var(value, 12, 1)
        if sched_entry:
            self._schedule['month']=sched_entry
            self._month = value
        else:
            raise exceptions.BadParametersException('Invalid month value: '+str(value))

    @property
    def dow(self):
        return self._dow

    @dow.setter
    def dow(self, value):
        sched_entry = self._process_var(value, 6, 0)
        if sched_entry:
            self._schedule['dow']=sched_entry
            self._dow = value
        else:
            raise exceptions.BadParametersException('Invalid dow value: '+str(value))

    @property
    def dom(self):
        return self._dom

    @dom.setter
    def dom(self, value):
        sched_entry = self._process_var(value, 31, 1)
        if sched_entry:
            self._schedule['dom']=sched_entry
            self._dom = value
        else:
            raise exceptions.BadParametersException('Invalid dom value: '+str(value))

    def _process_var(self, value, max_value, min_value):
        processed_entry=[]
        in_range=range(min_value,max_value+1)
        try:
            int_v=int(value)
            processed_entry.append(int_v) if (int_v<=max_value and int_v>=min_value) else None
        except Exception:
            if value=='*':
                for i in range(min_value,max_value+1):
                    processed_entry.append(i)
            elif len(value.split(','))>1:
                for group in value.split(','):
                    result=self._process_var(group,max_value,min_value)
                    for value in result:
                        processed_entry.append(value)
            elif len(value.split('-'))>1:
                r=value.split('-')
                r_min=int(r[0])
                r_max=int(r[1])
                for i in range(r_min,r_max+1):
                    if i in in_range:
                        processed_entry.append(i) 
            elif len(value.split('/'))>1:
                num,den=value.split('/')
                if num=='*':
                    num_list=in_range
                else:
                    num_int=int(num)
                    num_list=[num_int]
                if den=='*':
                    for i in num_list:
                        if i in in_range:
                            processed_entry.append(i)
                else:
                    den_int=int(den)
                    tmp_list=[]
                    for i in num_list:
                        if i%den_int==0:
                            tmp_list.append(i)
                    for j in tmp_list:
                        processed_entry.append(j)
        result_list=list(set(processed_entry))
        return result_list

    def meets(self, t):
        minute = t.tm_min
        hour = t.tm_hour
        month = t.tm_mon
        dow = t.tm_wday
        dom = t.tm_mday
        if (minute in self._schedule['minute'] and
           hour in self._schedule['hour'] and
           month in self._schedule['month'] and
           dow in self._schedule['dow'] and
           dom in self._schedule['dom']):
            return True
        return False




import unittest
import decimal
import pandas as pd
import datetime
import uuid
from komlogd.api.protocol.model import validation
from komlogd.api.protocol.model.types import Metrics

class ApiModelValidationTest(unittest.TestCase):

    def test_validate_uri_failure_invalid_uri(self):
        ''' validate_uri should fail if uri is not valid '''
        uris = [
            None,
            '  not valid ',
            '..not_valid',
            'not..valid',
            'ññ',
            'uri\n',
            'uri\t',
            'uri\r',
            121,
            323.1231,
            decimal.Decimal('23'),
            ['a','list'],
            {'set'},
            ('a','tuple'),
            {'a':'dict'},
        ]
        for uri in uris:
            with self.assertRaises(TypeError) as cm:
                validation.validate_uri(uri)

    def test_validate_uri_success(self):
        ''' validate_uri should return None if uri is valid '''
        uris = [
            'valid.uri',
            'valid.uri.-dash',
            'valid.uri._underscore',
            'valid',
        ]
        for uri in uris:
            self.assertTrue(validation.validate_uri(uri))

    def test_validate_ts_failure_invalid_ts(self):
        ''' validate_ts should fail if ts is not valid '''
        timestamps = [
            None,
            '  not valid ',
            '..not_valid',
            'not..valid',
            'ññ',
            'uri\n',
            'uri\t',
            'uri\r',
            '121',
            '323.1231',
            -1,
            -1.232,
            -1e2,
            2**32,
            decimal.Decimal('23'),
            decimal.Decimal(23),
            ['a','list'],
            {'set'},
            ('a','tuple'),
            {'a':'dict'},
            pd.Timestamp('now'),
            ['a','list'],
            ('a','tuple'),
            uuid.uuid1(),
            uuid.uuid4(),
            uuid.uuid4().hex,
            datetime.datetime.utcnow(),
            'NaN',
            'Infinity',
            '-Infinity',
            {'a':'dict'},
            {'set'},
            '3016-07-18T17:07:00.002323Z', #year greater than max supported
            '30160-07-18T17:07:00.002323Z', #year greater than max supported
            '-2016-07-18T17:07:00.002323Z', #negative years not allowed
            '1400-07-18T17:07:00.002323Z', #year lower than min supported
            '2016-7-18T17:07:00.002323Z', #month without leading 0
            '2016-07-38T17:07:00.002323Z', #day greater than 31
            '2016-07-8T17:07:00.002323Z', #day without leading 0
            '2016-07-18 17:07:00.002323Z', #no T Time separator
            '2016-07-1817:07:00.002323Z', #no Time separator
            '2016-07-18T24:07:00.002323Z', #hour 24 not valid
            '2016-07-18T17:60:00.002323Z', #min 60 not valid
            '2016-07-18T17:00:60.002323Z', #sec 60 not valid
            '2016-07-18T17:07:00.002323z', #z not capital
            '2016-07-18T17:07:00.002323+24:00',#24h tz offset not valid
            '2016-07-18T17:07:00.002323-23:60', #60min tz offset not valid
            '2016-07-18T17:07:00.002323-23:40:00', #seconds in tz offset not valid
            '2016-07-18T17:07:00.002323+0000', #no colon in offset not valid
            '2016/07/18T17:07:00.002323+0000', #no hyphen
            '2016-07-18T17:07:00.002323', #not tz dont allowed
            pd.Timestamp('now').isoformat(),
            datetime.datetime.utcnow().isoformat(),
        ]
        for ts in timestamps:
            with self.assertRaises(TypeError) as cm:
                validation.validate_ts(ts)

    def test_validate_ts_success(self):
        ''' validate_ts should return True if ts is valid '''
        timestamps = [
            pd.Timestamp('now',tz='utc'),
            pd.Timestamp('now',tz='Europe/Madrid'),
            '2016-07-18T17:07:00.002323Z',
            '2016-07-18T17:07:00.002323+00:22',
            '2016-07-18T17:07:00.002323-00:22',
            '2016-07-18T17:07:00+00:22',
        ]
        for ts in timestamps:
            self.assertTrue(validation.validate_ts(ts))

    def test_validate_ds_content_failure_invalid_content(self):
        ''' validate_ds_content should fail if content is not valid '''
        contents = [
            None,
            -1,
            -1.232,
            -1e2,
            2**32,
            decimal.Decimal('23'),
            decimal.Decimal(23),
            ['a','list'],
            {'set'},
            ('a','tuple'),
            {'a':'dict'},
            'a'*2**17+'a',
        ]
        for content in contents:
            with self.assertRaises(TypeError) as cm:
                validation.validate_ds_content(content)

    def test_validate_ds_content_success(self):
        ''' validate_ds_content should succeed if content is valid '''
        contents = [
            'a'*2**17,
            '§ºÆ¢¢§Ŧ®º¢Ð',
            'Some content\n with new lines\n tabs\t\tor ñ Ñ € æ¢ other chars.'
        ]
        for content in contents:
            self.assertTrue(validation.validate_ds_content(content))

    def test_validate_dp_content_failure_invalid_content(self):
        ''' validate_dp_content should fail if content is not valid '''
        contents = [
            None,
            ['a','list'],
            {'set'},
            ('a','tuple'),
            {'a':'dict'},
            'a'*2**17+'a',
            '1.1.1',
            '1,1',
            '1:1',
            '-1.e',
            '1e',
            '23,23e3',
            '23.23e3.3',
        ]
        for content in contents:
            with self.assertRaises(TypeError) as cm:
                validation.validate_dp_content(content)

    def test_validate_dp_content_success(self):
        ''' validate_dp_content should succeed if content is valid '''
        contents = [
            'NaN',
            decimal.Decimal('23'),
            decimal.Decimal(23),
            -1,
            -1.232,
            -1e2,
            2**32,
            '1'*2**7,
            '1.1',
            '1.5e+4',
            '1.005E+43',
            '-1.3e-34',
            '1e4',
            '-1e4',
            '-1.3e-3',
            '-1.3E-34',
            '1E4',
            '-1E4',
            '-1.3E-3',
            '+1.3E+3',
            '1.5e4\n',
            ' 23 ',
            ' 32\n',
            '32\n',
            '1'*2**7+'1',
        ]
        for content in contents:
            self.assertTrue(validation.validate_dp_content(content))

    def test_is_message_sequence_failure(self):
        ''' is_message_sequence should return False if param is no a valid sequence '''
        params = [
            'NaN',
            decimal.Decimal('23'),
            decimal.Decimal(23),
            -1,
            -1.232,
            -1e2,
            2**32,
            '1'*2**7,
            '1.1',
            '1.5e+4',
            '1.005E+43',
            '-1.3e-34',
            '1e4',
            '-1e4',
            '-1.3e-3',
            '-1.3E-34',
            '1E4',
            '-1E4',
            '-1.3E-3',
            '+1.3E+3',
            '1.5e4\n',
            ' 23 ',
            ' 32\n',
            '32\n',
            '1'*2**7+'1',
            uuid.uuid1(),
            uuid.uuid1().hex,
            uuid.uuid4().hex[0:20],
            uuid.uuid1().hex[0:10],
            uuid.uuid4(),
            {'set'},
            {'a':'dict'},
            ['a','list'],
            ('a','tuple'),
            None,
            1,
        ]
        for param in params:
            self.assertFalse(validation.is_message_sequence(param))

    def test_is_message_sequence_success(self):
        ''' is_message_sequence should return False if param is no a valid sequence '''
        params = [
            uuid.uuid1().hex[0:20],
        ]
        for param in params:
            self.assertTrue(validation.is_message_sequence(param))




import uuid
import decimal
import re
import pandas as pd
from datetime import datetime
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

LOCALURI=re.compile('^([a-zA-Z0-9\-_]+\.)*[a-zA-Z0-9\-_]+(?!\s)$')
GLOBALURI=re.compile('^([a-zA-Z0-9\-_]+\.)*[a-zA-Z0-9\-_]+:([a-zA-Z0-9\-_]+\.)*[a-zA-Z0-9\-_]+(?!\s)$')
USERNAME=re.compile('^([a-zA-Z0-9\-_]+\.)*[a-zA-Z0-9\-_]+(?!\s)$')
ISODATE=re.compile('^((?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$')

def validate_username(value):
    if isinstance(value,str) and USERNAME.search(value):
        return True
    raise TypeError('username is not valid: '+str(value))

def validate_uri(value):
    if not isinstance(value,str):
        raise TypeError('uri is not a string')
    if LOCALURI.search(value) or GLOBALURI.search(value):
        return True
    else:
        raise TypeError('uri is not valid: '+value)

def validate_ts(value):
    if ((isinstance(value, str) and ISODATE.search(value))
        or isinstance(value, pd.Timestamp)
        or isinstance(value, datetime)):
        try:
            t=pd.Timestamp(value)
            if t.tz is None:
                raise TypeError('timezone is required')
            if t.nanosecond != 0:
                raise TypeError('ts max precision')
            return True
        except ValueError:
            raise TypeError('ts value is out of bounds')
    else:
        raise TypeError('ts type is not valid')

def validate_privkey(value):
    if isinstance(value, RSAPrivateKey) and value.key_size >= 4096:
        return True
    raise TypeError('Invalid privkey')

def validate_ds_content(value):
    if not isinstance(value, str):
        raise TypeError('content is not a string')
    if len(value.encode('utf-8'))>2**17:
        raise TypeError('content size limit is 128K bytes')
    return True

def validate_dp_content(value):
    if isinstance(value,int) or isinstance(value,float):
        return True
    else:
        try:
            num=decimal.Decimal(str(value))
            return True
        except (decimal.InvalidOperation, ValueError, TypeError):
            raise TypeError('datapoint value not valid')

def is_message_sequence(value):
    if not (isinstance(value,str) and len(value)==20):
        return False
    try:
        s=uuid.UUID(value+'A'*12)
        if s.version == 1:
            return True
        else:
            return False
    except ValueError:
        return False

def is_local_uri(uri):
    if isinstance(uri,str) and LOCALURI.search(uri):
        return True
    return False

def is_global_uri(uri):
    if isinstance(uri,str) and GLOBALURI.search(uri):
        return True
    return False

def is_username(value):
    if isinstance(value,str) and USERNAME.search(value):
        return True
    return False




class BadParametersException(Exception):
    def __init__(self, msg=''):
        self.msg = msg

class InvalidMessageException(Exception):
    def __init__(self, msg=''):
        self.msg = msg

class LoginException(Exception):
    def __init__(self, msg=''):
        self.msg = msg

class WebsocketConnectionException(Exception):
    def __init__(self, msg=''):
        self.msg = msg

class CryptoException(Exception):
    def __init__(self, msg=''):
        self.msg = msg



'''

Transfer Methods

'''

import asyncio
import uuid
from komlogd.api import logging, uri
from komlogd.api.protocol.model import validation
from komlogd.api.protocol.model.schedules import OnUpdateSchedule, CronSchedule

class TransferMethodsIndex:

    def __init__(self, owner):
        self.owner = owner
        self._enabled_methods={}
        self._disabled_methods = {}
        self.uri_transfer_methods={}
        self.metrics={}
        self.on_update_uri_transfer_methods={}
        self.on_update_metrics={}

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        if value is None:
            self._owner = None
        else:
            validation.validate_username(value)
            self._owner = value.lower()

    def add_transfer_method(self, transfer_method, enabled=False):
        if self.get_transfer_method_info(transfer_method.mid):
            return False
        if self.owner:
            transfer_method._initialize(owner = self.owner)
            for m in transfer_method.m_in:
                guri = uri.get_global_uri(m, owner=self.owner)
                if guri not in self.metrics:
                    self.metrics[guri]=m
                try:
                    self.uri_transfer_methods[guri].add(transfer_method.mid)
                except KeyError:
                    self.uri_transfer_methods[guri]={transfer_method.mid}
            if isinstance(transfer_method.schedule, OnUpdateSchedule):
                for m in transfer_method.schedule.metrics:
                    guri = uri.get_global_uri(m, owner=self.owner)
                    if guri not in self.on_update_metrics:
                        self.on_update_metrics[guri]=m
                    try:
                        self.on_update_uri_transfer_methods[guri].add(transfer_method.mid)
                    except KeyError:
                        self.on_update_uri_transfer_methods[guri]={transfer_method.mid}
        if enabled:
            self._enabled_methods[transfer_method.mid]=transfer_method
        else:
            self._disabled_methods[transfer_method.mid]=transfer_method
        return True

    def enable_transfer_method(self, mid):
        if mid in self._disabled_methods:
            logging.logger.debug('enabling disabled method '+str(mid))
            self._enabled_methods[mid]=self._disabled_methods.pop(mid)
            return True
        elif mid in self._enabled_methods:
            logging.logger.debug('enabling already enabled method '+str(mid))
            return True
        else:
            return False

    def disable_transfer_method(self, mid):
        if mid in self._enabled_methods:
            logging.logger.debug('disabling enabled method '+str(mid))
            self._disabled_methods[mid]=self._enabled_methods.pop(mid)
            return True
        elif mid in self._disabled_methods:
            logging.logger.debug('disabling already disabled method '+str(mid))
            return True
        else:
            return False

    def delete_transfer_method(self, mid):
        self._enabled_methods.pop(mid,None)
        self._disabled_methods.pop(mid,None)
        for guri, mids in self.uri_transfer_methods.items():
            mids.pop(mid,None)
        return True

    def disable_all(self):
        for mid in list(self._enabled_methods.keys()):
            logging.logger.debug('Disabling transfer method '+str(mid))
            self._disabled_methods[mid]=self._enabled_methods.pop(mid)
        return True

    def get_transfer_method_info(self, mid):
        if mid in self._enabled_methods:
            return {'enabled':True, 'tm':self._enabled_methods[mid]}
        elif mid in self._disabled_methods:
            return {'enabled':False, 'tm':self._disabled_methods[mid]}
        else:
            return None

    def get_transfer_methods(self, metrics=None, enabled=True):
        mids=[]
        if metrics is None:
            if enabled:
                transfer_methods = list(self._enabled_methods.values())
            else:
                transfer_methods = list(self._disabled_methods.values())
        else:
            for m in metrics:
                if self.owner:
                    guri = uri.get_global_uri(m, owner=self.owner)
                else:
                    guri = m.uri
                if guri in self.metrics and self.metrics[guri].m_type is None:
                    self.metrics[guri]=m
                if guri in self.uri_transfer_methods:
                    for mid in self.uri_transfer_methods[guri]:
                        mids.append(mid)
            mids=list(set(mids))
            if enabled:
                transfer_methods=[self._enabled_methods[mid] for mid in mids if mid in self._enabled_methods]
            else:
                transfer_methods=[self._disabled_methods[mid] for mid in mids if mid in self._disabled_methods]
        return transfer_methods

    def get_on_update_transfer_methods(self, metrics):
        mids=[]
        for m in metrics:
            if self.owner:
                guri = uri.get_global_uri(m, owner=self.owner)
            else:
                guri = m.uri
            if guri in self.on_update_metrics and self.on_update_metrics[guri].m_type is None:
                self.on_update_metrics[guri]=m
            if guri in self.on_update_uri_transfer_methods:
                for mid in self.on_update_uri_transfer_methods[guri]:
                    mids.append(mid)
        mids=list(set(mids))
        transfer_methods=[self._enabled_methods[mid] for mid in mids if mid in self._enabled_methods]
        return transfer_methods

    def get_cron_transfer_methods(self, t):
        transfer_methods=[]
        for tm in self._enabled_methods.values():
            if isinstance(tm.schedule, CronSchedule) and tm.schedule.meets(t):
                transfer_methods.append(tm)
        return transfer_methods


anon_transfer_methods = TransferMethodsIndex(owner=None)



import uuid
import decimal
import pandas as pd
from komlogd.api.protocol.model import validation
from komlogd.api.protocol.model.types import Metrics, Actions, Datasource, Datapoint


class Catalog(type):
    def __init__(cls, name, bases, dct):
        if hasattr(cls, '_action_'):
            cls._catalog_[cls._action_.value]=cls
        super().__init__(name, bases, dct)

class KomlogMessage(metaclass=Catalog):
    _version_ = 1
    _catalog_ = {}

    def __new__(cls, *args, **kwargs):
        if cls is KomlogMessage:
            raise TypeError('<KomlogMessage> cannot be instantiated directly')
        return object.__new__(cls)

    @property
    def action(self):
        return self._action_

    @action.setter
    def action(self, value):
        raise TypeError('Action cannot be modified')

    @property
    def v(self):
        return self._version_

    @v.setter
    def v(self, value):
        raise TypeError('Version cannot be modified')

    @property
    def seq(self):
        return self._seq

    @seq.setter
    def seq(self, value):
        if hasattr(self, '_seq'):
            raise TypeError('Sequence cannot be modified')
        elif validation.is_message_sequence(value):
            self._seq = value
        else:
            raise TypeError('Invalid sequence')

    @property
    def irt(self):
        return self._irt

    @irt.setter
    def irt(self, value):
        if value is None or validation.is_message_sequence(value):
            self._irt = value
        else:
            raise TypeError('Invalid irt')

    @classmethod
    def load_from_dict(cls, msg):
        if cls is KomlogMessage:
            if isinstance(msg, dict) and 'action' in msg:
                try:
                    return cls._catalog_[msg['action']].load_from_dict(msg)
                except KeyError:
                    raise TypeError('Unknown message type')
            raise TypeError('Message not supported')
        else:
            raise NotImplementedError

class GenericResponse(KomlogMessage):
    _action_ = Actions.GENERIC_RESPONSE

    def __init__(self, status, error, reason, seq=None, irt=None):
        self.seq = seq if seq else uuid.uuid1().hex[0:20]
        self.irt = irt
        self.status = status
        self.error = error
        self.reason = reason

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        if isinstance(value, int) and value>0:
            self._status=value
        else:
            raise TypeError('Invalid status')

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, value):
        if isinstance(value, int) and value>=0:
            self._error=value
        else:
            raise TypeError('Invalid error')

    @property
    def reason(self):
        return self._reason

    @reason.setter
    def reason(self, value):
        if isinstance(value, str) or value is None:
            self._reason=value
        else:
            raise TypeError('Invalid reason')

    @classmethod
    def load_from_dict(cls, msg):
        if (isinstance(msg,dict)
            and 'v' in msg
            and 'action' in msg
            and 'seq' in msg
            and 'irt' in msg
            and 'payload' in msg
            and isinstance(msg['v'],int)
            and isinstance(msg['action'],str) and msg['action']==cls._action_.value
            and isinstance(msg['payload'],dict)
            and 'status' in msg['payload']
            and 'error' in msg['payload']
            and 'reason' in msg['payload']):
            status=msg['payload']['status']
            error=msg['payload']['error']
            reason=msg['payload']['reason']
            return cls(status=status, error=error, reason=reason, seq=msg['seq'], irt=msg['irt'])
        else:
            raise TypeError('Could not load message, invalid type')

class SendDsData(KomlogMessage):
    _action_ = Actions.SEND_DS_DATA

    def __init__(self, uri, ts, content, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt = irt
        self.uri=uri
        self.ts=ts
        self.content=content

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        validation.validate_uri(uri)
        self._uri=uri

    @property
    def ts(self):
        return self._ts

    @ts.setter
    def ts(self, ts):
        validation.validate_ts(ts)
        self._ts=pd.Timestamp(ts)

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        validation.validate_ds_content(content)
        self._content=content

    def to_dict(self):
        ''' returns a JSON serializable dict '''
        return {
            'v':self.v,
            'action':self.action.value,
            'seq':self.seq,
            'irt':self.irt,
            'payload':{
                'uri':self.uri,
                'ts':self.ts.isoformat(),
                'content':self.content
            }
        }

class SendDpData(KomlogMessage):
    _action_ = Actions.SEND_DP_DATA

    def __init__(self, uri, ts, content, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt=irt
        self.uri=uri
        self.ts=ts
        self.content=content

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        validation.validate_uri(uri)
        self._uri=uri

    @property
    def ts(self):
        return self._ts

    @ts.setter
    def ts(self, ts):
        validation.validate_ts(ts)
        self._ts=pd.Timestamp(ts)

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, content):
        validation.validate_dp_content(content)
        self._content = content if isinstance(content,decimal.Decimal) else decimal.Decimal(str(content))

    def to_dict(self):
        ''' returns a JSON serializable dict '''
        return {
            'v':self.v,
            'action':self.action.value,
            'seq':self.seq,
            'irt':self.irt,
            'payload':{
                'uri':self.uri,
                'ts':self.ts.isoformat(),
                'content':str(self.content)
            }
        }

class SendMultiData(KomlogMessage):
    _action_ = Actions.SEND_MULTI_DATA

    def __init__(self, ts, uris, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt=irt
        self.ts=ts
        self.uris=uris

    @property
    def uris(self):
        return self._uris

    @uris.setter
    def uris(self, uris):
        if (isinstance(uris,list)
            and all(isinstance(item,dict) for item in uris)
            and all('uri' in item for item in uris)
            and all(validation.validate_uri(item['uri']) for item in uris)
            and all('type' in item for item in uris)
            and all(item['type'] in [m.value for m in Metrics] + [m for m in Metrics] for item in uris)
            and all('content' in item for item in uris)
            and all(validation.validate_ds_content(item['content']) for item in uris if item['type'] in (Metrics.DATASOURCE.value, Metrics.DATASOURCE))
            and all(validation.validate_dp_content(item['content']) for item in uris if item['type'] in (Metrics.DATAPOINT.value, Metrics.DATAPOINT))):
            ds_uris=[{'uri':item['uri'],'type':Metrics(item['type']),'content':item['content']} for item in uris if item['type'] in ( Metrics.DATASOURCE.value, Metrics.DATASOURCE)]
            dp_uris=[{'uri':item['uri'],'type':Metrics(item['type']),'content':decimal.Decimal(str(item['content']))} for item in uris if item['type'] in (Metrics.DATAPOINT.value, Metrics.DATAPOINT)]
            self._uris=ds_uris+dp_uris
        else:
            raise TypeError('Uris parameter not valid')

    @property
    def ts(self):
        return self._ts

    @ts.setter
    def ts(self, ts):
        validation.validate_ts(ts)
        self._ts=pd.Timestamp(ts)

    @classmethod
    def load_from_dict(cls, msg):
        if (isinstance(msg, dict)
            and 'v' in msg
            and 'action' in msg
            and 'seq' in msg
            and 'irt' in msg
            and 'payload' in msg
            and isinstance(msg['v'],int) and msg['v']==cls._version_
            and isinstance(msg['action'],str) and msg['action']==cls._action_.value
            and isinstance(msg['payload'],dict)
            and 'ts' in msg['payload']
            and 'uris' in msg['payload']):
            ts=msg['payload']['ts']
            uris=msg['payload']['uris']
            return cls(ts=ts,uris=uris, seq=msg['seq'], irt=msg['irt'])
        else:
            raise TypeError('Could not load message, invalid type')

    def to_dict(self):
        ''' returns a JSON serializable dict '''
        ds_uris=[{'uri':item['uri'],'type':item['type'].value,'content':item['content']} for item in self._uris if item['type'] == Metrics.DATASOURCE]
        dp_uris=[{'uri':item['uri'],'type':item['type'].value,'content':str(item['content'])} for item in self._uris if item['type'] == Metrics.DATAPOINT]
        return {
            'v':self.v,
            'action':self.action.value,
            'seq':self.seq,
            'irt':self.irt,
            'payload':{
                'ts':self.ts.isoformat(),
                'uris':ds_uris+dp_uris
            }
        }

class HookToUri(KomlogMessage):
    _action_ = Actions.HOOK_TO_URI

    def __init__(self, uri, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt=irt
        self.uri=uri

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        validation.validate_uri(uri)
        self._uri=uri

    def to_dict(self):
        ''' returns a JSON serializable dict '''
        return {
            'v':self.v,
            'action':self.action.value,
            'seq':self.seq,
            'irt':self.irt,
            'payload':{
                'uri':self.uri
            }
        }

class UnHookFromUri(KomlogMessage):
    _action_ = Actions.UNHOOK_FROM_URI

    def __init__(self, uri, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt=irt
        self.uri = uri

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        validation.validate_uri(uri)
        self._uri=uri

    def to_dict(self):
        ''' returns a JSON serializable dict '''
        return {
            'v':self.v,
            'action':self.action.value,
            'seq':self.seq,
            'irt':self.irt,
            'payload':{
                'uri':self.uri
            }
        }

class RequestData(KomlogMessage):
    _action_ = Actions.REQUEST_DATA

    def __init__(self, uri, start=None, end=None, count=None, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt=irt
        self.uri = uri
        self.start = start
        self.end = end
        self.count = count

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        validation.validate_uri(uri)
        self._uri = uri

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        if start is not None:
            validation.validate_ts(start)
            self._start=pd.Timestamp(start)
        else:
            self._start = None

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        if end is not None:
            validation.validate_ts(end)
            self._end=pd.Timestamp(end)
        else:
            self._end = None

    @property
    def count(self):
        return self._count

    @count.setter
    def count(self, count):
        if isinstance(count,int) or count is None:
            self._count=count
        else:
            raise TypeError('Invalid count parameter')

    def to_dict(self):
        ''' returns a JSON serializable dict '''
        return {
            'v':self.v,
            'action':self.action.value,
            'seq':self.seq,
            'irt':self.irt,
            'payload':{
                'uri':self._uri,
                'start':self._start.isoformat() if self._start else None,
                'end':self._end.isoformat() if self._end else None,
                'count':self._count,
            }
        }

class SendDataInterval(KomlogMessage):
    _action_ = Actions.SEND_DATA_INTERVAL

    def __init__(self, metric, start, end, data, seq=None, irt=None):
        self.seq=seq if seq else uuid.uuid1().hex[0:20]
        self.irt=irt
        self.metric = metric
        self.start = start
        self.end = end
        self.data = data

    @property
    def metric(self):
        return self._metric

    @metric.setter
    def metric(self, metric):
        if isinstance(metric, Datasource) or isinstance(metric, Datapoint):
            self._metric = metric
        else:
            raise TypeError('Invalid metric type')

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        validation.validate_ts(start)
        self._start = pd.Timestamp(start)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        validation.validate_ts(end)
        self._end = pd.Timestamp(end)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if (isinstance(data, list)
            and all(
                isinstance(item,list)
                and len(item)==2
                and validation.validate_ts(item[0])
                and (validation.validate_dp_content(item[1]) if isinstance(self._metric, Datapoint) else validation.validate_ds_content(item[1]))
                for item in data)
            ):
            if isinstance(self._metric, Datasource):
                self._data=[(pd.Timestamp(row[0]),row[1]) for row in data]
            elif isinstance(self._metric, Datapoint):
                self._data=[(pd.Timestamp(row[0]),decimal.Decimal(str(row[1]))) for row in data]
        else:
            raise TypeError('Invalid data')

    @classmethod
    def load_from_dict(cls, msg):
        if (isinstance(msg,dict)
            and 'v' in msg
            and 'action' in msg
            and 'seq' in msg
            and 'irt' in msg
            and 'payload' in msg
            and isinstance(msg['v'],int) and msg['v']==cls._version_
            and isinstance(msg['action'],str) and msg['action']==cls._action_.value
            and isinstance(msg['payload'],dict)
            and 'uri' in msg['payload']
            and isinstance(msg['payload']['uri'],dict)
            and 'uri' in msg['payload']['uri']
            and 'type' in msg['payload']['uri']
            and 'start' in msg['payload']
            and 'end' in msg['payload']
            and 'data' in msg['payload']):
            if msg['payload']['uri']['type'] == Metrics.DATASOURCE.value:
                metric = Datasource(uri=msg['payload']['uri']['uri'])
            elif msg['payload']['uri']['type'] == Metrics.DATAPOINT.value:
                metric = Datapoint(uri=msg['payload']['uri']['uri'])
            else:
                raise TypeError ('Invalid metric type')
            start=msg['payload']['start']
            end=msg['payload']['end']
            data=msg['payload']['data']
            return cls(metric=metric,start=start,end=end,data=data,seq=msg['seq'],irt=msg['irt'])
        else:
            raise TypeError('Could not load message, invalid type')




import asyncio
import json
from komlogd.api import logging
from komlogd.api.protocol.processing import procedure as prproc
from komlogd.api.protocol.model.types import Metrics, Actions, Datasource, Datapoint, Metric


def process_message_send_multi_data(msg, session, **kwargs):
    metrics=[]
    for item in msg.uris:
        if item['type'] == Metrics.DATASOURCE:
            metric=Datasource(uri=item['uri'])
        elif item['type'] == Metrics.DATAPOINT:
            metric=Datapoint(uri=item['uri'])
        if not session._metrics_store.isin(metric=metric, ts=msg.ts, content=item['content']):
            session._metrics_store.store(metric, msg.ts, item['content'])
            metrics.append(metric)
    transfer_methods = session._transfer_methods.get_on_update_transfer_methods(metrics=metrics)
    for item in transfer_methods:
        logging.logger.debug('Requesting execution of method: '+item.f.__name__)
        asyncio.ensure_future(prproc.exec_transfer_method(mid=item.mid,ts=msg.ts,metrics=metrics,session=session))

def process_message_send_data_interval(msg, session, **kwargs):
    logging.logger.debug('Received data for uri: '+msg.metric.uri)
    logging.logger.debug('Interval ['+msg.start.isoformat()+' - '+msg.end.isoformat()+']')
    for row in msg.data[::-1]:
        session._metrics_store.store(metric=msg.metric, ts=row[0], content=row[1])

def process_message_generic_response(msg, session, **kwargs):
    logging.logger.debug('Received generic_response message')
    logging.logger.debug(str(msg.__dict__))

processing_map={
    Actions.SEND_MULTI_DATA:process_message_send_multi_data,
    Actions.SEND_DATA_INTERVAL:process_message_send_data_interval,
    Actions.GENERIC_RESPONSE:process_message_generic_response,
}



import os
from komlogd.api import crypto
from komlogd.base import config, exceptions, logging
from komlogd.base.settings import defaults, options

def get_private_key():
    keys=config.config.get_entries(entryname=options.KOMLOG_KEYFILE)
    if len(keys)==0:
        privkey_file=os.path.join(config.config.root_dir,defaults.RSA_PRIV_KEY)
    elif len(keys)>1:
        logging.logger.error('More than one keyfile entries found in configuration. Keep only one.')
        return None
    else:
        privkey_file=keys[0]
    if not os.path.isfile(privkey_file):
        logging.logger.debug('Generating RSA keys...')
        key_dir=os.path.split(privkey_file)[0]
        pubkey_file=os.path.join(key_dir,defaults.RSA_PUB_KEY)
        privkey=crypto.generate_rsa_key()
        if privkey and crypto.store_keys(privkey=privkey, privkey_file=privkey_file, pubkey_file=pubkey_file):
            logging.logger.debug('Keys stored successfully on disk')
            pubkey=privkey.public_key()
            key_str=crypto.get_printable_pubkey(pubkey)
            logging.logger.info('This is the public key, add it to your komlog account:\n'+key_str)
            return privkey
    else:
        privkey=crypto.load_private_key(privkey_file)
        return privkey

def get_public_key():
    privkey=get_private_key()
    if not privkey:
        logging.logger.error('Error obtaining public key.')
        return None
    else:
        pubkey=privkey.public_key()
        key_str=crypto.get_printable_pubkey(pubkey)
        logging.logger.info('Public key is:\n'+key_str)
        return pubkey

